import json.Field;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * @author dingweiran
 * @version 2022/6/22
 */
public class DataTypeParser {
    private static final Logger log = LoggerFactory.getLogger(DataTypeParser.class);

    public static final int MYSQL_TYPE_DECIMAL = 0;
    public static final int MYSQL_TYPE_INT8 = 1;
    public static final int MYSQL_TYPE_INT16 = 2;
    public static final int MYSQL_TYPE_INT32 = 3;
    public static final int MYSQL_TYPE_FLOAT = 4;
    public static final int MYSQL_TYPE_DOUBLE = 5;
    public static final int MYSQL_TYPE_NULL = 6;
    public static final int MYSQL_TYPE_TIMESTAMP = 7;
    public static final int MYSQL_TYPE_INT64 = 8;
    public static final int MYSQL_TYPE_INT24 = 9;
    public static final int MYSQL_TYPE_DATE = 10;
    public static final int MYSQL_TYPE_TIME = 11;
    public static final int MYSQL_TYPE_DATETIME = 12;
    public static final int MYSQL_TYPE_YEAR = 13;
    public static final int MYSQL_TYPE_DATE_NEW = 14;
    public static final int MYSQL_TYPE_VARCHAR = 15;
    public static final int MYSQL_TYPE_BIT = 16;
    public static final int MYSQL_TYPE_TIMESTAMP_NEW = 17;
    public static final int MYSQL_TYPE_DATETIME_NEW = 18;
    public static final int MYSQL_TYPE_TIME_NEW = 19;
    public static final int MYSQL_TYPE_JSON = 245;
    public static final int MYSQL_TYPE_DECIMAL_NEW = 246;
    public static final int MYSQL_TYPE_ENUM = 247;
    public static final int MYSQL_TYPE_SET = 248;
    public static final int MYSQL_TYPE_TINY_BLOB = 249;
    public static final int MYSQL_TYPE_MEDIUM_BLOB = 250;
    public static final int MYSQL_TYPE_LONG_BLOB = 251;
    public static final int MYSQL_TYPE_BLOB = 252;
    public static final int MYSQL_TYPE_VAR_STRING = 253;
    public static final int MYSQL_TYPE_STRING = 254;
    public static final int MYSQL_TYPE_GEOMETRY = 255;

    // 当mysql源库参数binlog_row_image设置为minimal时，
    // binlog中update记录的未修改字段的值是缺失的，
    // 这时用这个特定字符串代表缺失字段的值。
    public static final String JSON_NA_VALUE = "##DTS_NA_VALUE##";

    private static final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

    /**
     * @param data  可能有两种类型：string或null
     * @param field
     * @return
     */
    public static String decode(Object data, Field field) {
        if (data == null) {
            return "NULL";
        }
        String d = (String) data;

        if (JSON_NA_VALUE.equals(d)) {
            return "DEFAULT";
        }

        // 优先处理二进制类型。因为他们被处理成了hex string
        if (field.getOriginalType().toLowerCase().contains("binary")
                || field.getOriginalType().toLowerCase().contains("blob")
                || field.getDataTypeNumber() == MYSQL_TYPE_GEOMETRY) {
            return "x'" + d + "'";
        }
        switch (field.getDataTypeNumber()) {
            case MYSQL_TYPE_DECIMAL:
            case MYSQL_TYPE_DECIMAL_NEW:
            case MYSQL_TYPE_INT8:
            case MYSQL_TYPE_INT16:
            case MYSQL_TYPE_INT24:
            case MYSQL_TYPE_INT32:
            case MYSQL_TYPE_INT64:
            case MYSQL_TYPE_FLOAT:
            case MYSQL_TYPE_DOUBLE:
            case MYSQL_TYPE_BIT:
            case MYSQL_TYPE_NULL:
                return d;
            case MYSQL_TYPE_JSON:
                return "'" + d + "'";
            case MYSQL_TYPE_TIMESTAMP:
            case MYSQL_TYPE_TIMESTAMP_NEW: // 时间戳返回的是毫秒数
                long timestamp = Long.parseLong(d);
                return "_binary'" + formatter.format(new Date(timestamp)) + "'";
            case MYSQL_TYPE_DATETIME:
            case MYSQL_TYPE_DATETIME_NEW:
                String[] tstr = d.split(" ");
                if (tstr.length == 3) { // datetime 可能带时区信息
                    return "_binary'" + tstr[0] + " " + tstr[1] + "'";
                }
                return "_binary'" + d + "'";
            default:
                return "_binary'" + d + "'";
        }
    }
}
