import com.alibaba.fastjson.JSON;
import json.CanalFlatRecord;
import json.Const;
import json.Field;
import json.FlatRecord;
import org.apache.commons.cli.*;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;

import static json.Const.*;

public class SubscribeFromKafka {
    private static final Logger log = LoggerFactory.getLogger(SubscribeFromKafka.class);
    private static Map<Integer, partitionMsgConsumer> partitionMsgConsumers = new HashMap<>();

    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();
        Option option_canal = Option.builder()
                .required(false)
                .desc("The trans2canal option")
                .longOpt("trans2canal")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans)
                .addOption(option_canal);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql"),
                    commandLine.hasOption("trans2canal")
            );

        } catch (Exception exception) {
            log.info(exception.getMessage());
        }
    }

    private static void run(String brokers, String topic, String group, String user, String password, boolean trans2sql,
                            boolean trans2canal) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, 20 << 20); // kafka消息上限20M
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        KafkaConsumer<String, String> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(topic), new partitionSubLis());
        try {
            while (true) {
                ConsumerRecords<String, String> records = consumer.poll(Duration.ofSeconds(1));
                for (ConsumerRecord<String, String> record : records) {
                    partitionMsgConsumer pc = partitionMsgConsumers.get(record.partition());
                    if (null == pc) {
                        throw new IllegalStateException("no such partition consumer " + record.partition());
                    }

                    pc.consume(consumer, record, trans2sql, trans2canal);
                }
            }
        } finally {
            consumer.close();
        }
    }

    private final static String partitionSeqKey = "ps";
    private final static String partitionSeq2Key = "ps2";
    private final static String CHECKPOINT = "Checkpoint";
    private final static String HEARTBEAT = "Heartbeat";

    private static long getPartitionSeq(ConsumerRecord<String, String> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeqKey).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }

        return -1;
    }

    private static long getPartitionSeq2(ConsumerRecord<String, String> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeq2Key).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }
        return -1;
    }

    private static boolean isCheckpoint(ConsumerRecord<String, String> msg) {
        Iterator<Header> it = msg.headers().headers(CHECKPOINT).iterator();
        return it.hasNext();
    }

    private static boolean isHeartbeat(ConsumerRecord<String, String> msg) {
        Iterator<Header> it = msg.headers().headers(HEARTBEAT).iterator();
        return it.hasNext();
    }

    private static void onEntry(int partitionNum, long offset, long partitionSeq, FlatRecord record, boolean trans2sql,
                                boolean trans2canal) {
        if (!trans2canal) {
            log.info("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                    + ", " + JSON.toJSONString(record));
        } else {
            // trans to canal flat json format
            CanalFlatRecord canalRecord = trans2canal(record);
            log.info("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                    + ", " + JSON.toJSONString(canalRecord));
        }

        if (trans2sql) {
            try {
                trans2SQL(partitionNum, offset, partitionSeq, record);
            } catch (Throwable e) {
                log.error(e.getMessage());
            }
        }
    }

    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, FlatRecord record) {
        if (!Const.HEARTBEAT.equals(record.getMessageType()) && !Const.CHECKPOINT.equals(record.getMessageType())) {
            log.info(
                    String.format("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%s], happenedAt: %s",
                            partitionNum,
                            offset,
                            partitionSeq,
                            record.getFileName(), record.getPosition(),
                            LocalDateTime.ofInstant(Instant.ofEpochSecond(record.getTimestamp()),
                                    TimeZone.getDefault().toZoneId())));
        }

        switch (record.getMessageType()) {
            case Const.HEARTBEAT:
                log.info("heartbeat: " + record.getHeartbeatTimestamp());
                break;
            case Const.CHECKPOINT:
                log.info("Checkpoint: gtid=" + record.getSyncedGtid()
                        + ", file=" + record.getFileName() + ", position=" + record.getPosition());
                break;
            case Const.BEGIN:
                log.info("BEGIN");
                break;
            case Const.COMMIT:
                log.info("COMMIT");
                break;
            case Const.ROLLBACK:
                log.info("ROLLBACK");
                break;
            case DDL:
            case INIT_DDL:
            case STATEMENT:
                log.info(record.getSql());
                break;
            case Const.INSERT:
                log.info(transInsert(record));
                break;
            case Const.INIT_INSERT:
                log.info(transInitInsert(record));
                break;
            case Const.DELETE:
                log.info(transDelete(record));
                break;
            case Const.UPDATE:
                log.info(transUpdate(record));
                break;
            default:
                throw new IllegalStateException("unsupported message type");
        }
    }

    private static String transInsert(FlatRecord record) {
        StringBuilder sql = new StringBuilder(String.format(
                "INSERT INTO `%s`.`%s` VALUES ",
                record.getSchemaName(),
                record.getTableName()));
        // 该版本类型只有 List<String>
        List<String> row = (List) record.getNewColumns();
        List<Field> columns = record.getColumns();
        String sep = "";
        sql.append('(');
        for (int i = 0; i < row.size(); i++) {
            sql.append(sep);
            sql.append(DataTypeParser.decode(row.get(i), columns.get(i)));
            sep = ", ";
        }
        sql.append(')');
        return sql.toString();
    }

    private static String transInitInsert(FlatRecord record) {
        List<Field> columns = record.getColumns();
        List<String> colNames = new ArrayList<>();
        for (Field col : columns) {
            colNames.add(col.getName());
        }
        StringBuilder sql = new StringBuilder(String.format(
                "INSERT INTO `%s`.`%s`(%s) VALUES ",
                record.getSchemaName(),
                record.getTableName(),
                String.join(",", colNames)));
        // 该版本类型只有 List<String>
        List<String> row = (List) record.getNewColumns();
        String sep = "";
        sql.append('(');
        for (int i = 0; i < row.size(); i++) {
            sql.append(sep);
            sql.append(DataTypeParser.decode(row.get(i), columns.get(i)));
            sep = ", ";
        }
        sql.append(')');
        return sql.toString();
    }

    private static String transDelete(FlatRecord record) {
        StringBuilder sql = new StringBuilder(String.format(
                "DELETE FROM `%s`.`%s` WHERE ",
                record.getSchemaName(),
                record.getTableName()));
        String sep = "";
        List<Field> columns = record.getColumns();
        List row = (List) record.getOldColumns();
        for (int i = 0; i < row.size(); i++) {
            Field column = columns.get(i);
            Object data = row.get(i);
            String d = DataTypeParser.decode(data, column);
            if ("DEFAULT".equals(d) || "NULL".equals(d)) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s` = %s", column.getName(), d));

            sep = " AND ";
        }
        sql.append(" LIMIT 1");
        return sql.toString();
    }

    private static String transUpdate(FlatRecord record) {
        StringBuilder sql = new StringBuilder(String.format(
                "UPDATE `%s`.`%s` SET ",
                record.getSchemaName(),
                record.getTableName()));

        String sep = "";
        List<Field> columns = record.getColumns();
        List newRow = (List) record.getNewColumns();
        for (int i = 0; i < newRow.size(); i++) {
            Field column = columns.get(i);
            Object data = newRow.get(i);
            if (DataTypeParser.JSON_NA_VALUE.equals(data)) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s` = %s", column.getName(), DataTypeParser.decode(data, column)));
            sep = ", ";
        }

        List oldRow = (List) record.getOldColumns();
        sql.append(" WHERE ");
        sep = "";
        for (int i = 0; i < oldRow.size(); i++) {
            Field column = columns.get(i);
            Object data = oldRow.get(i);
            String d = DataTypeParser.decode(data, column);
            if ("DEFAULT".equals(d) || "NULL".equals(d)) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s` = %s", column.getName(), d));
            sep = " AND ";
        }
        sql.append(" LIMIT 1");
        return sql.toString();
    }

    private static class partitionMsgConsumer {
        long lastPS = 0;
        long lastPS2 = 0;
        TopicPartition tp;

        private partitionMsgConsumer(TopicPartition tp) {
            this.tp = tp;
        }

        private void consume(KafkaConsumer<String, String> consumer, ConsumerRecord<String, String> record, boolean trans2sql, boolean trans2canal) {
            // Each message has a sequence id in the partition, this id is continuous monotonic increasing,
            // these messages are duplicated if the id rotate;
            // There is unexpected message lost in kafka if the id is discontinuous,
            // have to reset the subscribe and client when it happened.
            // ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
            // ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
            // 重复解析部分binlog导致的数据重复投递问题。
            long ps2 = getPartitionSeq2(record);
            long validPs;
            if (ps2 != -1) {
                if (lastPS2 > 0) {
                    if (ps2 < lastPS2) {
                        System.out.println("Duplicated message: " + record);
                        return;
                    } else if (ps2 == lastPS2) { // 可能是重复消息也可能是checkpoint或heartbeat消息
                        if (!isCheckpoint(record) && !isHeartbeat(record)) {
                            System.out.println("Duplicated message: " + record);
                            return;
                        }
                    } else if (ps2 != lastPS2 + 1) {
                        throw new IllegalStateException("the partition seq2 is discontinuous, last: " + lastPS2 + ", current: " + ps2);
                    }
                }
                lastPS2 = ps2;
                validPs = ps2;
            } else { // 老的订阅任务没有ps2，会走到这里
                long ps = getPartitionSeq(record);
                if (ps == -1) {

                } else if (ps <= lastPS) {
                    System.out.println("Duplicated message: " + record);
                    return;
                } else if (ps != lastPS + 1 && lastPS != 0) {
                    throw new IllegalStateException("the partition seq is discontinuous, last: " + lastPS + ", current: " + ps);
                }
                lastPS = ps;
                validPs = ps;
            }

            FlatRecord recordValue = Deserializer.parse(record.value());

            onEntry(record.partition(), record.offset(), validPs, recordValue, trans2sql, trans2canal);

            // 提交位点
            if (Const.CHECKPOINT.equals(recordValue.getMessageType())) {
                consumer.commitAsync(Collections.singletonMap(tp, new OffsetAndMetadata(record.offset() + 1)), null);
            }
        }
    }

    private static class partitionSubLis implements ConsumerRebalanceListener {
        public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.computeIfAbsent(p.partition(), k -> new partitionMsgConsumer(p));
            }
        }

        public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.remove(p.partition());
            }
        }
    }

    private static CanalFlatRecord trans2canal(FlatRecord record) {
        CanalFlatRecord canalRecord = new CanalFlatRecord(record.getId());
        canalRecord.setDatabase(record.getSchemaName());
        canalRecord.setTable(record.getTableName());
        canalRecord.setPkNames(record.getPkNames());
        canalRecord.setIsDdl(DDL.equals(record.getMessageType()));
        canalRecord.setType(record.getMessageType());
        canalRecord.setEs((long) record.getExecutionTime());
        canalRecord.setTs(record.getReaderTimestamp());
        canalRecord.setSql(record.getSql());
        if (!INSERT.equals(record.getMessageType()) && !UPDATE.equals(record.getMessageType())
                && !DELETE.equals(record.getMessageType())) {
            return canalRecord;
        }

        Map<String, Integer> sqlType = new LinkedHashMap<>();
        Map<String, String> mysqlType = new LinkedHashMap<>();
        canalRecord.setSqlType(sqlType);
        canalRecord.setMysqlType(mysqlType);
        for (Field field : record.getColumns()) {
            sqlType.put(field.getName(), field.getDataTypeNumber());
            mysqlType.put(field.getName(), field.getOriginalType());
        }

        List<Map<String, String>> canalData = new ArrayList<>();
        List<Map<String, String>> canalOld = new ArrayList<>();
        canalRecord.setData(canalData);
        canalRecord.setOld(canalOld);

        List<Field> columns = record.getColumns();
        // 如果是DELETE，也把删除前的数据放到data字段，为了兼容canal的设计
        List dataRows;
        if (DELETE.equals(record.getMessageType())) {
            dataRows = (List) record.getOldColumns();
        } else {
            dataRows = (List) record.getNewColumns();
        }
        // 判断数据字段个数是否和表头字段个数一致
        if (columns.size() != dataRows.size()) {
            log.error("columns size not equal data field size, something really bad happened!");
            return canalRecord;
        }
        for (int i = 0; i < dataRows.size(); i++) {
            Field column = columns.get(i);
            String row = (String) dataRows.get(i);
            Map<String, String> canalRow = new LinkedHashMap<>();
            canalRow.put(column.getName(), row);
            canalData.add(canalRow);
        }
        if (UPDATE.equals(record.getMessageType())) {
            // 只有UPDATE会保存old字段
            List oldRows = (List) record.getOldColumns();
            if (columns.size() != oldRows.size()) {
                log.error("columns size not equal data field size, something really bad happened!");
                return canalRecord;
            }
            for (int i = 0; i < oldRows.size(); i++) {
                Field column = columns.get(i);
                String oldRow = (String) oldRows.get(i);
                Map<String, String> canalRow = new LinkedHashMap<>();
                canalRow.put(column.getName(), oldRow);
                canalOld.add(canalRow);
            }
        }
        return canalRecord;
    }
}